/*

*/

import tree.*;
import tokenizer.*;
import treelanguage.*;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.List;

/**
   This program implements an "interpreter" for the tree language.
   <p>
   This program provides both a REPL (Read Eval Print Loop)
   for the language and a "script file" interpreter.
*/
public class Tree_Language
{
   public static void main(String[] args) throws IOException
   {
      if (args.length > 1)
      {
         System.out.println("Usage: Tree_Language <script>");
      }
      else if (args.length == 1)
      {
         runFile(args[0]);
      }
      else
      {
         runPrompt();
      }
   }


   private static void runFile(final String path) throws IOException
   {
      final byte[] bytes = Files.readAllBytes(Paths.get(path));

      final String source = new String(bytes, Charset.defaultCharset());

      // Tokenize, parse, and evaluate the source file.
      try
      {
         final Tokenizer tokens = new Tokenizer(source);

         // Print the token list.
         System.out.println(tokens);

         final Tree<String> tree = Parse2Tree.parse(tokens);

         // Print the Tree as an S-expression
         System.out.println( tree + "\n" );

         // Pretty-print the Tree.
         System.out.println( PrettyPrinter.prettyPrint( tree ) + "\n" );

         // Create dot and png files from the Tree.
         final String baseName = path;
         Tree2dot.tree2dot(tree, baseName);
         Tree2png.tree2png(baseName);
      }
      catch (TokenizeError e)
      {
         System.out.println(e);
      }
      catch (ParseError e)
      {
         System.out.println(e);
      }
   }


   private static void runPrompt() throws IOException
   {
      final InputStreamReader input = new InputStreamReader(System.in);
      final BufferedReader reader = new BufferedReader(input);

      System.out.println("Tree_Lang.");

      // REPL (read, eval, print loop).
      for (int i = 1;  ; ++i)
      {
         System.out.print("> ");

         // Read.
         final String source = reader.readLine();

         try
         {
            final Tokenizer tokens = new Tokenizer(source);
            if ( ! tokens.hasToken() )
            {
               continue; // skip over empty string
            }
            final Tree<String> tree = Parse2Tree.parse(tokens);

            if ( tokens.hasToken() )  // there shouldn't be any more tokens
               throw new ParseError("unexpected input: "
                                  + tokens.peek().lexeme
                                  + " at line " + tokens.peek().line
                                  + ", position " + tokens.peek().position
                                   );

            // Pretty-print the tree.
            System.out.println( PrettyPrinter.prettyPrint( tree ) + "\n" );

            // Create dot, png, and XML files from the Tree.
            final String baseName = String.format("repl.%02d", i);
            Tree2dot.tree2dot(tree, baseName);
            Tree2png.tree2png(baseName);
            Tree2xml.tree2xml(tree, baseName);
         }
         catch (TokenizeError e)
         {
            System.out.println(e);
         }
         catch (ParseError e)
         {
            System.out.println(e);
         }
      }
   }
}
