/*

*/

package tokenizer;

import java.util.List;
import java.util.ArrayList;

/**
   This {@code Tokenizer} splits a {@link String}
   into a {@link List} of {@link Token} objects.
   <p>
   This tokenizer is split into two parts.
   <p>
   The first part is a private part that tokenizes
   a given input {@link String} into a {@link List}
   of {@link Token} objects. The private part of this
   tokenizer is only used by the {@code Tokenizer}
   constructor.
   <p>
   The second part of this tokenizer is a public part
   that lets a client program parse the {@link List}
   of {@link Token} objects that was built by the
   private part (the constructor).
   <p>
   This tokenizer is adapted from the online book
   "Crafting Interpreters," by Robert Nystrom.
   <p>
   See <a href="https://craftinginterpreters.com/scanning.html" target="_top">
                https://craftinginterpreters.com/scanning.html</a><br>
   and <a href="https://github.com/munificent/craftinginterpreters/blob/master/java/com/craftinginterpreters/lox/Scanner.java" target="_top">
   https://github.com/munificent/craftinginterpreters/blob/master/java/com/craftinginterpreters/lox/Scanner.java</a><br>
*/
public class Tokenizer
{
   // tokenizing state (used by the constructor)
   private final String source;
   private int currentChar;
   private int startLexeme;
   private int line = 1;
   private int startOfLine = currentChar;

   // tokenizer state (used by the public interface)
   private final List<Token> tokens = new ArrayList<>();
   private int currentToken;

   /**
      This constructor has the private part of
      this file build the list of tokens from the
      given input string.

      @param source  {@link String} to be tokenized
      @throws TokenizeError if there is an error in the string
   */
   public Tokenizer(final String source)
   {
      this.source = source;
      scanTokens();
      currentToken = 0;
   }

   /**
      Break up the input string into individual tokens.
   */
   private void scanTokens()
   {
      while (! isAtEnd())
      {
         // We are at the beginning of the next lexeme.
         startLexeme = currentChar;
         scanToken(); // consume one token from the input string
      }
   }

   /**
      Consume one token and add it to the list of tokens.
   */
   private void scanToken()
   {
      final char c = advance();
      switch (c)
      {
         // one character tokens
         case '(': addToken(); break;
         case ')': addToken(); break;
         case ',': addToken(); break;
         case ';': addToken(); break;
         case '%': addToken(); break;
         case '^': addToken(); break;

         // one or two character tokens
         case '*': match('*'); addToken(); break;
         case '|': match('|'); addToken(); break;
         case '&': match('&'); addToken(); break;
         case '=': match('='); addToken(); break;
         case '!': match('='); addToken(); break;
         case '<': match('='); addToken(); break;
         case '>': match('='); addToken(); break;

         // division (possible single line comment)
         case '/':
           if (match('/'))
           {
              // A comment goes until the end of the line.
              while (peekChar() != '\n' && !isAtEnd()) advance();
           }
           else
           {
              addToken();
           }
           break;

         // plus (possible positive integer or increment operator)
         case '+':
           if (isDigit(peekChar())) // positive integer
           {
              advance();
              integer();
           }
           else // addition operator or increment operator
           {
              match('+'); addToken();
           }
           break;

         // minus (possible negative integer or decrement operator)
         case '-':
           if (isDigit(peekChar())) // negative integer
           {
              advance();
              integer();
           }
           else // subtraction operator or decrement operator
           {
              match('-'); addToken();
           }
           break;

         // whitespace
         case ' ':
         case '\r':
         case '\t':
           // Ignore whitespace.
           break;
         case '\n':
           ++line; // Count line numbers.
           startOfLine = currentChar;
           break;

         default:
           if (isDigit(c)) // digit start
           {
              integer();
           }
           else if (isAlpha(c)) // identifier start
           {
              variable();
           }
           else // error
           {
              throw new TokenizeError("unexpected character: "
                                    + "'" + c + "'"
                                    + " at line " + line
                                    + ", position " + (startLexeme-startOfLine+1)
                                    + "\n");
           }
           break;
      }
   }

   private void addToken()
   {
      final String text = source.substring(startLexeme, currentChar);
      tokens.add(new Token(text, line, startLexeme - startOfLine + 1));
   }

   private char advance()
   {
      ++currentChar;
      return source.charAt(currentChar - 1);
   }

   private boolean match(final char expected)
   {
      if (isAtEnd()) return false;
      if (source.charAt(currentChar) != expected) return false;
      ++currentChar;
      return true;
   }

   private char peekChar()
   {
      if (isAtEnd()) return '\0';
      return source.charAt(currentChar);
   }

   private char peek2Char()
   {
      if (currentChar + 1 >= source.length()) return '\0';
      return source.charAt(currentChar + 1);
   }

   private boolean isAtEnd()
   {
      return currentChar >= source.length();
   }

   private void variable()
   {
      while (isAlphaNumeric(peekChar()))
         advance();

      addToken();
   }

   private void integer()
   {
      while (isDigit(peekChar()))
         advance();

      addToken();
   }

   private boolean isAlpha(final char c)
   {
      return (c >= 'a' && c <= 'z') ||
             (c >= 'A' && c <= 'Z') ||
              c == '_';
   }

   private boolean isAlphaNumeric(final char c)
   {
      return isAlpha(c) || isDigit(c);
   }

   private boolean isDigit(final char c)
   {
      return c >= '0' && c <= '9';
   }


   // The following methods are the public interface to the tokenizer.

   /**
      @return true if all of the tokens have been consumed
   */
   public boolean isEmpty()
   {
      return currentToken >= tokens.size();
   }

   /**
      @return true if there are tokens that have not been consumed
   */
   public boolean hasToken()
   {
      return currentToken < tokens.size();
   }

   /**
      @return the number of unconsumed tokens
   */
   public int size()
   {
      return tokens.size() - currentToken;
   }

   /**
      This method consumes and returns the current token.

      @return the next {@link Token}
   */
   public Token next()
   {
      return tokens.get(currentToken++);
   }

   /**
      This method both tests and consumes the current token.

      @param lexeme {@link String} to be compared with the next {@link Token}
      @return true if {@code lexeme} matches the next {@link Token}
   */
   public boolean match(final String lexeme)
   {
      final boolean match = lexeme.equals( tokens.get(currentToken).lexeme );
      if (match)
      {
         ++currentToken; // consume the matched token
         return true;
      }
      else
      {
         return false; // don't consume the unmatched token
      }
   }

   /**
      This method allows you to look at the current
      token without consuming it.

      @return the current {@link Token} without consuming it
   */
   public Token peek()
   {
      return tokens.get(currentToken);
   }

   /**
      This method allows you to look at the token after
      the current token (without consuming any token).

      @return the {@link Token} after the current {@link Token}
   */
   public Token peek2()
   {
      return tokens.get(currentToken + 1);
   }

   @Override
   public String toString()
   {
      String result = "";
      for (final Token t : tokens)
      {
         result += t + "\n";
      }
      return result;
   }
}
