/*

*/

package tree;

import java.util.List;
import java.util.LinkedList;

/**
   This class defines a tree data structure
   as an object that has a reference to a
   root object and a reference to a list of
   sub trees.
*/
public class LinkedTree<T> implements Tree<T>
{
   private final T element;
   private final List<Tree<T>> subTrees;

   /**
      Construct a leaf node.

      @param element  reference to the data object to store in this node
      @throws NullPointerException if {@code element} is {@code null}
   */
   public LinkedTree(final T element)
   {
      if (null == element)
         throw new NullPointerException("root element must not be null");

      this.element = element;
      this.subTrees = new LinkedList<>();
   }


   /**
      Construct a {@code LinkedTree} with the given {@link List} of
      {@link Tree}s as its branches.

      @param element  reference to the data object to store in this node
      @param subTrees  {@link List} of {@link Tree} branches for this node
      @throws NullPointerException if {@code element} is {@code null}
      @throws NullPointerException if {@code subTrees} is {@code null}
   */
   public LinkedTree(final T element, final List<Tree<T>> subTrees)
   {
      if (null == element)
         throw new NullPointerException("root element must not be null");
      if (null == subTrees)
         throw new NullPointerException("subTrees must not be null");

      this.element = element;
      this.subTrees = subTrees;
   }


   /**
      Construct a {@code LinkedTree} with the given {@link Tree}s
      as its branches.

      @param element   reference to the data object to store in this node
      @param subTrees  array of {@link Tree} branches for this node
      @throws NullPointerException if {@code element} is {@code null}
   */
   @SafeVarargs // maybe?
   public LinkedTree(final T element, final Tree<T>... subTrees) // variable number of arguments
   {
      if (null == element)
         throw new NullPointerException("root element must not be null");

      this.element = element;
      this.subTrees = new LinkedList<Tree<T>>();
      for (final Tree<T> tree : subTrees)
      {
         this.subTrees.add(tree);
      }
   }


   /**
      This is a static factory method.

      Convert an arbitrary {@link Tree} to a {@code LinkedTree}.

      @param <T>   The element type for the {@link Tree}
      @param tree  A {@link Tree} of any type
      @return a {@code LinkedTree} version of the input tree.
      @throws NullPointerException if {@code tree} is {@code null}
   */
   public static <T> LinkedTree<T> convert(final Tree<T> tree)
   {
      if (null == tree)
         throw new NullPointerException("tree must not be null");

      final List<Tree<T>> subTrees = new LinkedList<>();

      for ( final Tree<T> subTree : tree.subTrees() )
      {
         subTrees.add( convert(subTree) );  // recursion
      }

      return new LinkedTree<T>(tree.root(), subTrees);
   }


   // Implement the two methods of the Tree<T> interface.

   @Override
   public T root()
   {
      return element;
   }

   @Override
   public List<Tree<T>> subTrees()
   {
      return subTrees;
   }

   /**
      Convert a {@code TreeA} into a {@link String}.

      @return a {@link String} representation of this {@code TreeA}
   */
   @Override
   public String toString()
   {
      String result = "";
      if ( isLeaf() )
      {
         result += root();
      }
      else
      {
         result += "(" + root();
         for ( final Tree<T> tree : subTrees() )
         {
            result += " " + tree;  // recursion
         }
         result += ")";
      }
      return result;
   }
}
