/*

*/

package treelanguage;

import tree.*;
import tokenizer.*;

import java.io.PrintWriter;
import java.io.File;

/**
   This program converts a {@link Tree} of {@link Token}s
   into an XML language description of the {@link Tree}.
   The XML description is written to a {@link File} with
   the .xml extension. The .xml file can then be opened
   with an XML viewer (for example, any web browser) to
   produce a graphical representation of the {@link Tree}
   data structure.
<p>
   See
   <br>
   <a href="https://en.wikipedia.org/wiki/XML" target="_top">
            https://en.wikipedia.org/wiki/XML</a>
*/

public class AST2xml
{
   /**
      Use the {@link String} {@code baseName} as the base for
      the .xml file that this program produces from the
      {@link Tree} of {@link Token}s.

      @param ast       {@link Tree} of {@link Token}s to convert into an xml file
      @param baseName  the base name for the created xml file
   */
   public static void ast2xml(final Tree<Token> ast, final String baseName)
   {
      // Create an .xml file from a tree.
      try
      {
         // Create an empty .xml file.
         final PrintWriter out = new PrintWriter(
                                    new File(baseName + ".xml"));
         // Put xml code into the .xml file.
         out.println("<?xml version=\"1.0\" ?>");
         out.println("<AbstractSyntaxTree>");
         // Traverse the tree and convert it to appropriate xml tags.
         out.println( ast2xml(ast) );
         out.println("</AbstractSyntaxTree>");
         out.close();
      }
      catch (Exception e)
      {
         System.out.println( e );
      }
   }


   /**
      Return a {@link String} representation of the
      xml tags that describes the given {@link Tree}
      of {@link Token}s.
      <p>
      This method is essentially a pre-order traversal
      of the tree.

      @param ast  {@link Tree} of {@link Token}s to convert into an xml {@link String}
      @return a {@link String} of xml tags representing the abstract syntax tree
   */
   public static String ast2xml(final Tree<Token> ast)
   {
      String result = "<tree>";
      result += "<root>";
      final String lexeme = ast.root().lexeme;
      if ( lexeme.equals("<") )
      {
         result += "&lt;";
      }
      else if ( lexeme.equals("<=") )
      {
         result += "&lt;=";
      }
      else if ( lexeme.equals(">") )
      {
         result += "&gt;";
      }
      else if ( lexeme.equals(">=") )
      {
         result += "&gt;=";
      }
      else if ( lexeme.equals("&&") )
      {
         result += "&amp;&amp;";
      }
      else
      {
         result += lexeme;
      }
      result += "</root>";
      // Recursively traverse all the sub trees.
      for ( final Tree<Token> subTree : ast.subTrees() )
      {
         // Convert the sub tree into an xml description.
         result += ast2xml(subTree);
      }
      result += "</tree>";

      return result;
   }
}
